use std::net::SocketAddr;

use log::{error, info};
use quick_xml::{
    events::{attributes::Attributes, BytesDecl, BytesEnd, BytesStart, Event},
    Reader, Writer,
};
use tokio::{
    io::{AsyncWrite, AsyncWriteExt, BufReader},
    net::{
        tcp::{ReadHalf, WriteHalf},
        TcpStream,
    },
};

use crate::{
    error::StreamError,
    feature::Feature,
    negotiator::{self, Step},
    tag::{self, Tag},
};

type Result<T> = std::result::Result<T, StreamError>;

const FEATURES: &'static [Feature] = &[Feature::start_tls(true)];

struct StreamStart<'a> {
    reader: Reader<BufReader<ReadHalf<'a>>>,
    writer: Writer<WriteHalf<'a>>,
    buffer: Vec<u8>,
    hostname: String,
}

impl<'a> StreamStart<'a> {
    fn new(stream: &'a mut TcpStream, hostname: String) -> Self {
        let (read, write) = stream.split();
        let (reader, writer) = (
            Reader::from_reader(BufReader::new(read)),
            Writer::new(write),
        );

        Self {
            reader,
            writer,
            hostname,
            buffer: vec![],
        }
    }

    async fn start_stream(mut self) {
        match self.negotiate_stream().await {
            Ok(_) => {}
            Err(err) => {
                if let Err(err2) = error(self.writer.get_mut(), err).await {
                    error!("error writing error: {err2}");
                    return;
                } else {
                    info!("wrote error {err}")
                }

                if let Err(e) = self.writer.get_mut().write_all(b"</stream:stream>").await {
                    error!("writing end to stream: {e}")
                }
                if let Err(e) = self.writer.get_mut().shutdown().await {
                    error!("shutting down stream: {e}")
                }
            }
        }
    }

    async fn negotiate_stream(&mut self) -> Result<()> {
        let attrs = loop {
            match self.reader.read_event_into_async(&mut self.buffer).await? {
                Event::Start(start) => {
                    if start.name().as_ref() == tag::STREAM_ELEMENT_NAME {
                        let attrs: StreamAttrs = start.attributes().try_into()?;
                        if attrs.namespace != XMLNamespace::JabberClient {
                            return Err(StreamError::InvalidNamespace);
                        }
                        break attrs;
                    } else {
                        info!("element: {:?}", start);
                    }
                }
                Event::End(_) => return Err(StreamError::BadFormat),
                Event::Eof => return Err(StreamError::BadFormat),
                _ => continue,
            }
        };
        info!("starting negotiation with: {attrs:?}");
        self.write_stream_header(StreamAttrs {
            from: attrs.to.clone(),
            to: attrs.from,
            namespace: XMLNamespace::JabberClient,
        })
        .await?;
        if attrs.to != self.hostname {
            return Err(StreamError::HostUnknown);
        }
        self.send_features().await?;
        loop {
            match self.reader.read_event_into_async(&mut self.buffer).await? {
                Event::Empty(empty) => match empty.name().as_ref() {
                    tag::STARTTLS => {
                        info!("starttls negotiation");
                        if let Step::Failure = negotiator::start_tls(
                            self.reader.get_mut(),
                            self.writer.get_mut(),
                            empty,
                        )
                        .await?
                        {
                            return Step::Failure.write_tag(self.writer.get_mut()).await;
                        };
                    }
                    _ => return Err(StreamError::UnsupportedFeature),
                },
                Event::End(_) => return Err(StreamError::BadFormat),
                Event::Eof => return Err(StreamError::BadFormat),
                _ => continue,
            }
        }

        Err(StreamError::InternalServerError)
    }
    async fn write_stream_header(&mut self, req: StreamAttrs) -> Result<()> {
        self.writer
            .write_event_async(Event::Decl(BytesDecl::new("1.0", Some("utf-8"), None)))
            .await?;
        self.writer
            .write_event_async(Event::Start(
                BytesStart::new("stream:stream").with_attributes(vec![
                    ("from", req.from.as_str()),
                    ("to", req.to.as_str()),
                    ("xmlns:stream", "http://etherx.jabber.org/streams"),
                    ("xml:lang", "en"),
                    ("version", "1.0"),
                ]),
            ))
            .await?;

        Ok(())
    }

    async fn send_features(&mut self) -> Result<()> {
        self.writer
            .write_event_async(Event::Start(BytesStart::new(tag::FEATURE)))
            .await?;
        for feature in FEATURES.into_iter() {
            feature.write_tag(self.writer.get_mut()).await?;
        }
        self.writer
            .write_event_async(Event::End(BytesEnd::new(tag::FEATURE)))
            .await?;
        Ok(())
    }
}

pub fn spawn(hostname: String, (mut stream, _): (TcpStream, SocketAddr)) {
    tokio::spawn(async move {
        StreamStart::new(&mut stream, hostname).start_stream().await;
    });
}

async fn error<W: AsyncWrite + Unpin>(writer: W, err: StreamError) -> Result<()> {
    let mut writer = Writer::new(writer);
    let err = err.to_string();
    writer
        .write_event_async(Event::Start(BytesStart::new(tag::ERROR_ELEMENT)))
        .await?;
    writer
        .write_event_async(Event::Start(
            BytesStart::new(&err)
                .with_attributes(vec![("xmlns", "urn:ietf:params:xml:ns:xmpp-streams")]),
        ))
        .await?;
    writer
        .write_event_async(Event::End(BytesEnd::new(&err)))
        .await?;
    writer
        .write_event_async(Event::End(BytesEnd::new(tag::ERROR_ELEMENT)))
        .await?;
    Ok(())
}

#[derive(Debug, Clone)]
struct StreamAttrs {
    from: String,
    to: String,
    namespace: XMLNamespace,
}

impl TryFrom<Attributes<'_>> for StreamAttrs {
    type Error = StreamError;

    fn try_from(value: Attributes<'_>) -> std::result::Result<Self, Self::Error> {
        let mut from: Option<String> = None;
        let mut to: Option<String> = None;
        let mut ns: Option<XMLNamespace> = None;
        for v in value {
            let v = v?;
            match v.key.local_name().into_inner() {
                b"from" => {
                    from = Some(String::from_utf8(v.value.to_vec())?);
                }
                b"to" => {
                    to = Some(String::from_utf8(v.value.to_vec())?);
                }
                b"xmlns" => match v.value.to_vec().as_slice() {
                    b"jabber:client" => {
                        ns = Some(XMLNamespace::JabberClient);
                    }
                    _ => return Err(StreamError::InvalidNamespace),
                },
                other => {
                    info!(
                        "ignoring key {}",
                        String::from_utf8(other.to_vec()).unwrap_or_default()
                    );
                }
            }
        }
        Ok(StreamAttrs {
            from: from.ok_or(StreamError::InvalidFrom)?,
            to: to.ok_or(StreamError::HostUnknown)?,
            namespace: ns.ok_or(StreamError::BadNamespacePrefix)?,
        })
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum XMLNamespace {
    JabberClient,
}
